from dataclasses import dataclass
from datetime import datetime
from enum import Enum

import freighter

from arya import telem, channel
from arya.channel import entity
from arya.segment import BinarySegment, NumpySegment, encoder

_ENDPOINT = "/segment/iterate"


class Command(int, Enum):
    OPEN = 0
    NEXT = 1
    PREV = 2
    FIRST = 3
    LAST = 4
    NEXT_SPAN = 5
    PREV_SPAN = 6
    NEXT_RANGE = 7
    VALID = 8
    ERROR = 9
    SEEK_FIRST = 11
    SEEK_LAST = 12
    SEEK_LT = 13
    SEEK_GE = 14


class ResponseVariant(int, Enum):
    ACK = 1
    DATA = 2


@dataclass
class Request:
    command: Command
    span: telem.TimeSpan | None = None
    range: telem.TimeRange | None = None
    stamp: telem.TimeStamp | None = None
    keys: list[str] | None = None


@dataclass
class Response:
    variant: ResponseVariant
    ack: bool
    command: Command
    error: freighter.ErrorPayload
    segments: list[BinarySegment]

    def load(self, data: dict):
        if data["variant"] == 0:
            return
        self.variant = ResponseVariant(data["variant"])
        if self.variant == ResponseVariant.ACK:
            self.ack = data["ack"]
            self.command = Command(data["command"])
            if "error" in data:
                self.error = freighter.ErrorPayload(data["error"]["type"],
                                                    data["error"]["data"])
        elif self.variant == ResponseVariant.DATA:
            self.segments = [BinarySegment(**seg) for seg in data["segments"]]
        else:
            raise ValueError("Unexpected response variant")


def _response_factory() -> Response:
    return Response(ResponseVariant.ACK, False, Command.OPEN,
                    freighter.ErrorPayload(None, None), [])


class Core:
    transport: freighter.StreamClient
    stream: freighter.Stream[Request, Response]
    values: list[BinarySegment]
    aggregate: bool

    def __init__(self, transport: freighter.StreamClient, aggregate: bool = False) -> None:
        self.transport = transport
        self.aggregate = aggregate

    def open(self, keys: list[str], tr: telem.TimeRange):
        self.stream = self.transport.stream(_ENDPOINT, Request, _response_factory)
        self.exec(command=Command.OPEN, range=tr, keys=keys)
        self.values = []

    def exec(self, **kwargs) -> bool:
        exc = self.stream.send(Request(**kwargs))
        if exc is not None:
            raise exc
        if not self.aggregate:
            self.values = []
        while True:
            r, exc = self.stream.receive()
            if exc is not None:
                raise exc
            if r.variant == ResponseVariant.ACK:
                return r.ack
            self.values += r.segments

    def next(self) -> bool:
        return self.exec(command=Command.NEXT)

    def prev(self) -> bool:
        return self.exec(command=Command.PREV)

    def first(self) -> bool:
        return self.exec(command=Command.FIRST)

    def last(self) -> bool:
        return self.exec(command=Command.LAST)

    def next_span(self, span: telem.TimeSpan) -> bool:
        return self.exec(command=Command.NEXT_SPAN, span=span)

    def prev_span(self, span: telem.TimeSpan) -> bool:
        return self.exec(command=Command.PREV_SPAN, span=span)

    def next_range(self, rng: telem.TimeRange) -> bool:
        return self.exec(command=Command.NEXT_RANGE, range=rng)

    def seek_first(self) -> bool:
        return self.exec(command=Command.SEEK_FIRST)

    def seek_last(self) -> bool:
        return self.exec(command=Command.SEEK_LAST)

    def seek_lt(self, stamp: telem.TimeStamp) -> bool:
        return self.exec(command=Command.SEEK_LT, stamp=stamp)

    def seek_ge(self, stamp: telem.TimeStamp) -> bool:
        return self.exec(command=Command.SEEK_GE, stamp=stamp)

    def valid(self) -> bool:
        return self.exec(command=Command.VALID)

    def exhaust(self) -> bool:
        return self.exec(command=Command.EXHAUST)

    def close(self):
        exc = self.stream.close_send()
        if exc is not None:
            raise exc
        pld, exc = self.stream.receive()
        print(pld)
        if not isinstance(exc, freighter.EOF):
            print(exc)
            raise exc


class Numpy(Core):
    decoder: encoder.NumpyEncoderDecoder
    channel_client: channel.Client
    channels: channel.Registry

    def __init__(
            self,
            transport: freighter.StreamClient,
            channel_client: channel.Client,
            aggregate: bool = False
    ):
        super().__init__(transport, aggregate)
        self.decoder = encoder.NumpyEncoderDecoder()
        self.channel_client = channel_client

    def open(self, keys: list[str], tr: telem.TimeRange) -> None:
        self.channels = channel.Registry(self.channel_client.retrieve(keys))
        super().open(keys, tr)

    @property
    def value(self) -> dict[str, NumpySegment]:
        decoded = []
        self.values.sort(key=lambda v: v.start)
        res = dict()
        for i, seg in enumerate(self.values):
            decoded.append(self.decoder.decode(self.channels.get(seg.channel_key), seg))
        for i, dec in enumerate(decoded):
            if dec.channel.key in res:
                res[dec.channel.key].extend(dec)
            else:
                res[dec.channel.key] = dec
        return res



from .entity import SugaredBinarySegment
from .. import telem


class Splitter:
    threshold: telem.Size

    def __init__(self, threshold: telem.Size) -> None:
        self.threshold = threshold

    def split(self, segment: SugaredBinarySegment) -> list[SugaredBinarySegment]:
        if segment.size <= self.threshold:
            return [segment]
        else:
            split_v = self.threshold - (self.threshold % segment.channel.density)
            truncated = SugaredBinarySegment(
                segment.channel, segment.start, segment.data[: split_v]
            )
            _next = SugaredBinarySegment(
                segment.channel, truncated.end, segment.data[split_v:]
            )
            return [truncated, *self.split(_next)]



from freighter import Endpoint, StreamClient, AsyncStreamClient, UnaryClient
from freighter import (
    ws,
    sync,
    http,
    encoder,
)


class Transport:
    endpoint: Endpoint
    stream: StreamClient
    stream_async: AsyncStreamClient
    http: http.Client

    def __init__(self, endpoint: Endpoint) -> None:
        self.endpoint = endpoint.child("/api/v1/")
        self.stream_async = ws.Client(endpoint=self.endpoint, encoder=encoder.Msgpack(), max_message_size=int(5e6))
        self.stream = sync.StreamClient(self.stream_async)
        self.http = http.Client(endpoint=self.endpoint, encoder_decoder=encoder.JSON())



from arya.channel import Channel


class Registry:
    channels: dict[str, Channel]

    def __init__(self, channels: list[Channel]) -> None:
        self.channels = {ch.key: ch for ch in channels}

    def get(self, key: str) -> Channel | None:
        return self.channels.get(key, None)



from __future__ import annotations

import datetime
from dataclasses import dataclass
import numpy as np
import pandas as pd
from typing import get_args
from typing import Union

import arya.errors

_EPOCH = datetime.datetime.utcfromtimestamp(0)


class TimeStamp(int):
    """TimeStamp represents a 64 bit nanosecond-precision UTC timestamp.

    :param value: The nanosecond value of the timestamp
    """

    def __new__(cls, value: UnparsedTimeStamp, *args, **kwargs):
        if isinstance(value, TimeStamp):
            return value

        if isinstance(value, TimeSpan):
            value = int(value)
        elif isinstance(value, pd.Timestamp):
            value = int(value.asm8.view(np.int64))
        elif isinstance(value, datetime.datetime):
            value = SECOND * int((value - _EPOCH).total_seconds())
        elif isinstance(value, datetime.timedelta):
            value = SECOND * int(value.total_seconds())
        elif isinstance(value, np.datetime64):
            value = int(pd.Timestamp(value).asm8.view(np.int64))
        elif isinstance(value, int):
            return super().__new__(cls, int(value))
        else:
            raise TypeError(f"Cannot convert {type(value)} to TimeStamp")

        return super().__new__(cls, value)

    def __init__(self, value: UnparsedTimeStamp, *args, **kwargs):
        pass

    def time(self) -> datetime.datetime:
        """Returns the TimeStamp represented as a datetime.datetime object.
        :return: a datetime.datetime object
        """
        return datetime.datetime.utcfromtimestamp(self / SECOND)

    def is_zero(self) -> bool:
        """Returns true if the TimeStamp is zero.
        :return: True if the TimeStamp is zero, False otherwise
        """
        return self == 0

    def after(self, ts: UnparsedTimeStamp) -> bool:
        """Returns true if the TimeStamp is after the given TimeStamp.
        :param ts: the TimeStamp to compare to
        :return: True if the TimeStamp is after the given TimeStamp, False otherwise
        """
        return super().__gt__(TimeStamp(ts))

    def after_eq(self, ts: UnparsedTimeStamp) -> bool:
        """Returns true if the TimeStamp is after or equal to the given TimeStamp.
        :param ts: the TimeStamp to compare to
        :return: True if the TimeStamp is after or equal to the given TimeStamp an d False
        otherwise.
        """
        return super().__ge__(TimeStamp(ts))

    def before(self, ts: UnparsedTimeStamp) -> bool:
        """Returns true if the TimeStamp is before the given TimeStamp.
        :param ts: the TimeStamp to compare to
        :return: True if the TimeStamp is before the given TimeStamp, False otherwise
        """
        return super().__lt__(TimeStamp(ts))

    def before_eq(self, ts: UnparsedTimeStamp) -> bool:
        """Returns true if the TimeStamp is before or equal to the given TimeStamp.
        :param ts: the TimeStamp to compare to
        :return: True if the TimeStamp is before or equal to the given TimeStamp, and False
        otherwise.
        """
        return super().__le__(TimeStamp(ts))

    def span_range(self, span: TimeSpan) -> TimeRange:
        """Returns a TimeRange that spans the given TimeSpan.
        :param span: the TimeSpan to span
        :return: a TimeRange that spans the given TimeSpan
        """
        rng = TimeRange(self, self + span)
        if not rng.is_valid():
            rng = rng.swap()
        return rng

    def range(self, ts: UnparsedTimeStamp) -> TimeRange:
        """Returns a new TimeRange spanning the provided time stamps
        :param ts: the second time stamp
        :return: a new TimeRange spanning the provided time stamps
        """
        return TimeRange(self, TimeStamp(ts))

    def add(self, ts: UnparsedTimeStamp) -> TimeStamp:
        """Returns a new TimeStamp that is the sum of the two TimeStamps.
        :param ts: the second TimeStamp
        :return: a new TimeStamp that is the sum of the two TimeStamps
        """
        return TimeStamp(super().__add__(TimeStamp(ts)))

    def sub(self, ts: UnparsedTimeStamp) -> TimeStamp:
        """Returns a new TimeStamp that is the difference of the two TimeStamps.
        :param ts: the second TimeStamp
        :return: a new TimeStamp that is the difference of the two TimeStamps
        """
        return TimeStamp(super().__sub__(TimeStamp(ts)))

    def __add__(self, rhs: UnparsedTimeStamp) -> TimeStamp:
        return self.add(rhs)

    def __sub__(self, rhs: UnparsedTimeStamp) -> TimeStamp:
        return self.sub(rhs)

    def __lt__(self, rhs: UnparsedTimeStamp) -> bool:
        return self.before(rhs)

    def __le__(self, rhs: UnparsedTimeStamp) -> bool:
        return self.before_eq(rhs)

    def __ge__(self, rhs: UnparsedTimeStamp) -> bool:
        return self.after_eq(rhs)

    def __gt__(self, rhs: UnparsedTimeStamp) -> bool:
        return self.after(rhs)

    def __eq__(self, rhs: object) -> bool:
        if isinstance(rhs, get_args(UnparsedTimeStamp)):
            return super().__eq__(TimeStamp(rhs))
        return NotImplemented


def now() -> TimeStamp:
    return TimeStamp(datetime.datetime.now())


class TimeSpan(int):
    """TimeSpan represents a 64 bit nanosecond-precision duration."""

    def __new__(cls, value: UnparsedTimeSpan, *args, **kwargs):
        if isinstance(value, int):
            return super().__new__(cls, value)
        elif isinstance(value, TimeSpan):
            return value

        if isinstance(value, datetime.timedelta):
            value = int(float(SECOND) * value.total_seconds())
        return super().__new__(cls, value)

    def __init__(self, value: UnparsedTimeSpan, *args, **kwargs):
        pass

    def delta(self) -> datetime.timedelta:
        """Returns the TimeSpan represented as a datetime.timedelta object.
        :return: a datetime.timedelta object
        """
        return datetime.timedelta(seconds=self.seconds())

    def seconds(self) -> float:
        """Returns the TimeSpan represented as a number of seconds.
        :return: a number of seconds
        """
        return float(self / SECOND)

    def is_zero(self) -> bool:
        """Returns true if the TimeSpan is zero.
        :return: True if the TimeSpan is zero, False otherwise
        """
        return self == 0

    def add(self, ts: UnparsedTimeSpan) -> TimeSpan:
        """Returns a new TimeSpan that is the sum of the two TimeSpans.
        :param ts: the second TimeSpan
        :return: a new TimeSpan that is the sum of the two TimeSpans
        """
        return TimeSpan(super().__add__(TimeSpan(ts)))

    def sub(self, ts: UnparsedTimeSpan) -> TimeSpan:
        """Returns a new TimeSpan that is the difference of the two TimeSpans.
        :param ts: the second TimeSpan
        :return: a new TimeSpan that is the difference of the two TimeSpans
        """
        return TimeSpan(super().__sub__(TimeSpan(ts)))

    def __add__(self, other: UnparsedTimeSpan) -> TimeSpan:
        return self.add(other)

    def __sub__(self, other: UnparsedTimeSpan) -> TimeSpan:
        return self.sub(other)

    def __mul__(self, other: UnparsedTimeSpan) -> TimeSpan:
        return TimeSpan(super().__mul__(TimeSpan(other)))

    def __rmul__(self, other: UnparsedTimeSpan) -> TimeSpan:
        return self.__mul__(other)

    def __gt__(self, other: UnparsedTimeSpan) -> bool:
        return super().__gt__(TimeSpan(other))

    def __ge__(self, other: UnparsedTimeSpan) -> bool:
        return super().__ge__(TimeSpan(other))

    def __lt__(self, other: UnparsedTimeSpan) -> bool:
        return super().__lt__(TimeSpan(other))

    def __le__(self, other: UnparsedTimeSpan) -> bool:
        return super().__le__(TimeSpan(other))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, get_args(UnparsedTimeSpan)):
            return NotImplemented
        return super().__eq__(TimeSpan(other))


TIME_STAMP_MIN = TimeStamp(0)
TIME_STAMP_MAX = TimeStamp(0xFFFFFFFFFFFFFFFF)
NANOSECOND = TimeSpan(1)
MICROSECOND = TimeSpan(1000) * NANOSECOND
MILLISECOND = TimeSpan(1000) * MICROSECOND
SECOND = TimeSpan(1000) * MILLISECOND
MINUTE = TimeSpan(60) * SECOND
HOUR = TimeSpan(60) * MINUTE
TIME_SPAN_MAX = TimeSpan(0xFFFFFFFFFFFFFFFF)


class Rate(float):
    """Rate represents a data rate in Hz"""

    def __new__(cls, value: UnparsedRate):
        if isinstance(value, float):
            return super().__new__(cls, value)
        if isinstance(value, Rate):
            return value
        if isinstance(value, TimeSpan):
            value = 1 / value.seconds()
        elif isinstance(value, int):
            value = float(value)
        else:
            raise TypeError(f"Cannot convert {type(value)} to Rate")
        return super().__new__(cls, value)

    def __init__(self, value: UnparsedRate):
        pass

    def period(self) -> TimeSpan:
        """Returns the period of the rate as a TimeSpan"""
        return TimeSpan(int(1 / self * float(SECOND)))

    def sample_count(self, time_span: UnparsedTimeSpan) -> int:
        """Returns the number of samples in the given TimeSpan at this rate"""
        return int(TimeSpan(time_span) / self.period())

    def byte_size(self, time_span: UnparsedTimeSpan, density: Density) -> Size:
        """Calculates the amount of bytes occupied by the given TimeSpan at the given rate and sample density."""
        return Size(self.sample_count(time_span) * int(density))

    def span(self, sample_count: int) -> TimeSpan:
        """Returns the TimeSpan that corresponds to the given number of samples at this rate."""
        return self.period() * sample_count

    def size_span(self, size: Size, density: Density) -> TimeSpan:
        """Returns the TimeSpan that corresponds to the given number of bytes at this rate and sample density."""
        if size % density != 0:
            raise arya.errors.ContiguityError(
                f"Size {size} is not a multiple of density {density}"
            )
        return self.span(int(size / density))

    def __str__(self):
        return str(int(self)) + "Hz"

    def __repr__(self):
        return str(int(self)) + "Hz"


HZ = Rate(1)
KHZ = Rate(1000) * HZ
MHZ = Rate(1000) * KHZ


@dataclass
class TimeRange:
    start: TimeStamp
    end: TimeStamp

    def __init__(self, start: UnparsedTimeStamp, end: UnparsedTimeStamp):
        self.start = TimeStamp(start)
        self.end = TimeStamp(end)

    def span(self) -> TimeSpan:
        return TimeSpan(self.end - self.start)

    def is_zero(self) -> bool:
        return self.span().is_zero()

    def bound_by(self, otr: TimeRange) -> TimeRange:
        if otr.start.after(self.start):
            self.start = otr.start
        if otr.start.after(self.end):
            self.end = otr.start
        if otr.end.before(self.end):
            self.end = otr.end
        if otr.end.before(self.start):
            self.start = otr.end
        return self

    def contains_stamp(self, ts: TimeStamp) -> bool:
        return ts.after_eq(self.start) and ts.before(self.end)

    def contains_range(self, tr: TimeRange) -> bool:
        return self.start.before_eq(tr.start) and self.end.after_eq(tr.end)

    def overlaps_with(self, tr: TimeRange) -> bool:
        return (
                self.contains_stamp(tr.start)
                or self.contains_stamp(tr.end)
                or tr.contains_range(self)
        )

    def swap(self) -> TimeRange:
        self.start, self.end = self.end, self.start
        return self

    def is_valid(self) -> bool:
        return self.span() >= TimeSpan(0)

    def __str__(self):
        return str(self.start) + " - " + str(self.end)


class Density(int):
    ...

    def __new__(cls, value: UnparsedDensity):
        if isinstance(value, Density):
            return value
        if isinstance(value, int):
            return super().__new__(cls, value)
        raise TypeError(f"Cannot convert {type(value)} to Density")

    def __init__(self, value):
        pass


class Size(int):
    def __str__(self):
        return super(Size, self).__str__() + "B"


TIME_RANGE_MAX = TimeRange(TIME_STAMP_MIN, TIME_STAMP_MAX)
DENSITY_UNKNOWN = Density(0)
BIT64 = Density(8)
BIT32 = Density(4)
BIT16 = Density(2)
BIT8 = Density(1)


class DataType(str):

    def __new__(cls, value: UnparsedDataType):
        if isinstance(value, DataType):
            return value
        if isinstance(value, str):
            return super().__new__(cls, value)
        try:
            if issubclass(value, np.ScalarType):
                return from_numpy_type(value)
        except TypeError:
            pass
        raise TypeError(f"Cannot convert {type(value)} to DataType")

    def __init__(self, value: UnparsedDataType):
        pass


def to_numpy_type(data_type: DataType) -> np.ScalarType:
    return NUMPY_TYPES.get(data_type, None)


def from_numpy_type(np_type: np.ScalarType) -> DataType:
    return DATA_TYPES.get(np_type, None)


DATA_TYPE_UNKNOWN = DataType("")
FLOAT64 = DataType("float64")
FLOAT32 = DataType("float32")
INT64 = DataType("int64")
INT32 = DataType("int32")
INT16 = DataType("int16")
INT8 = DataType("int8")
UINT64 = DataType("uint64")
UINT32 = DataType("uint32")
UINT16 = DataType("uint16")
UINT8 = DataType("uint8")

UnparsedTimeStamp = Union[
    TimeStamp,
    TimeSpan,
    int,
    datetime.datetime,
    datetime.timedelta,
]
UnparsedTimeSpan = Union[TimeSpan | TimeStamp | int | datetime.timedelta]
UnparsedRate = TimeSpan | Rate | float
UnparsedDensity = Density | int
UnparsedDataType = (*np.ScalarType, DataType, str)

NUMPY_TYPES: dict[str, np.ScalarType] = {
    FLOAT64: np.float64,
    FLOAT32: np.float32,
    INT64: np.int64,
    INT32: np.int32,
    INT16: np.int16,
    INT8: np.int8,
    UINT64: np.uint64,
    UINT32: np.uint32,
    UINT16: np.uint16,
    UINT8: np.uint8,
}

DATA_TYPES: dict[np.ScalarType, DataType] = {
    np.float64: FLOAT64,
    np.float32: FLOAT32,
    np.int64: INT64,
    np.int32: INT32,
    np.int16: INT16,
    np.int8: INT8,
    np.uint64: UINT64,
    np.uint32: UINT32,
    np.uint16: UINT16,
    np.uint8: UINT8,
}



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, TypeAlias

import freighter
from .entity import Channel
from .. import telem
from ..transport import Transport


@dataclass
class _RetrieveRequest:
    keys: Optional[list[str]] = None
    node_id: int = None
    names: Optional[list[str]] = None


@dataclass
class _Response:
    channels: list[Channel]

    def load(self, data: dict):
        self.channels = [Channel(**c) for c in data["channels"]]


def _response_factory() -> _Response:
    return _Response(channels=[])


@dataclass
class _CreateRequest:
    channel: Channel
    count: int


_RETRIEVE_ENDPOINT = "/channel/retrieve"
_CREATE_ENDPOINT = "/channel/create"

ChannelRetrieveTransport: TypeAlias = freighter.UnaryClient[
    _RetrieveRequest,
    _Response,
]
ChannelCreateTransport: TypeAlias = freighter.UnaryClient[_CreateRequest, _Response]


class Client:
    retrieve_transport: ChannelRetrieveTransport
    create_transport: ChannelCreateTransport

    def __init__(self, transport: Transport):
        self.retrieve_transport = transport.http.get(
            _RetrieveRequest,
            _response_factory,
        )
        self.create_transport = transport.http.post(
            _CreateRequest,
            _response_factory,
        )

    def retrieve(self, keys: list[str]) -> list[Channel]:
        return self._retrieve(_RetrieveRequest(keys=keys))

    def retrieve_by_name(self, names: list[str]) -> list[Channel]:
        return self._retrieve(_RetrieveRequest(names=names))

    def retrieve_by_node_id(self, node_id: int) -> list[Channel]:
        return self._retrieve(_RetrieveRequest(node_id=node_id))

    def create(
            self,
            name: str = "",
            node_id: int = 0,
            rate: telem.UnparsedRate = telem.Rate(0),
            data_type: telem.UnparsedDataType = telem.DATA_TYPE_UNKNOWN,
    ) -> Channel:
        return self.create_n(Channel(name, node_id, rate, data_type), 1)[0]

    def create_n(self, channel: Channel, count: int = 1) -> list[Channel]:
        req = _CreateRequest(channel=channel, count=count)
        res, exc = self.create_transport.send(_CREATE_ENDPOINT, req)
        if exc is not None:
            raise exc
        assert res is not None
        return res.channels

    def _retrieve(self, req: _RetrieveRequest) -> list[Channel]:
        res, exc = self.retrieve_transport.send(_RETRIEVE_ENDPOINT, req)
        if exc is not None:
            raise exc
        assert res is not None
        return res.channels



from __future__ import annotations
from dataclasses import dataclass
from functools import reduce

import numpy
import numpy as np

from arya import telem
from arya.channel import Channel
from arya.telem import TimeStamp, TimeSpan, TimeRange


@dataclass
class SugaredHeader:
    channel: Channel
    start: TimeStamp = TimeStamp(0)

    def __init__(self, channel: Channel, start: telem.UnparsedTimeStamp):
        self.channel = channel
        self.start = TimeStamp(start)


@dataclass
class Header:
    channel_key: str
    start: TimeStamp = TimeStamp(0)

    def __init__(self, channel_key: str, start: telem.UnparsedTimeStamp):
        self.channel_key = channel_key
        self.start = TimeStamp(start)

    def sugar(self, channel: Channel) -> SugaredHeader:
        return SugaredHeader(channel, self.start)


@dataclass
class BinarySegment(Header):
    data: bytes = b""

    def __init__(
            self,
            channel_key: str = "",
            start: telem.UnparsedTimeStamp = telem.TimeStamp(0),
            data: bytes = b"",
    ):
        super().__init__(channel_key, start)
        self.data = data

    def sugar(self, channel: Channel) -> SugaredBinarySegment:
        return SugaredBinarySegment(channel, self.start, self.data)

    @property
    def size(self) -> telem.Size:
        return telem.Size(len(self.data))


@dataclass
class SugaredBinarySegment(SugaredHeader):
    data: bytes = b""

    def __init__(
            self, channel: Channel, start: telem.UnparsedTimeStamp, data: bytes = b""
    ):
        super().__init__(channel, start)
        self.data = data

    def desugar(self) -> BinarySegment:
        return BinarySegment(self.channel.key, self.start, self.data)

    @property
    def size(self) -> telem.Size:
        return telem.Size(len(self.data))

    @property
    def span(self) -> TimeSpan:
        return self.channel.rate.size_span(self.size, self.channel.density)

    @property
    def range(self) -> TimeRange:
        return self.start.span_range(self.span)

    @property
    def end(self) -> TimeStamp:
        return self.range.end


@dataclass
class NumpySegment(SugaredHeader):
    data: numpy.ndarray = numpy.array([])

    def __init__(self, channel: Channel, start: TimeStamp, data: numpy.ndarray):
        super().__init__(channel, start)
        self.data = data

    @property
    def span(self) -> TimeSpan:
        return self.channel.rate.span(self.data.size)

    @property
    def range(self) -> TimeRange:
        return self.start.span_range(self.span)

    @property
    def end(self) -> TimeStamp:
        return self.range.end

    def extend(self, other: NumpySegment):
        assert self.channel == other.channel
        assert self.end == other.start
        self.data = np.append(self.data, other.data)



from datetime import datetime

import numpy as np

from .. import channel, telem
from ..transport import Transport
from . import writer
from . import iterator
from .entity import NumpySegment


class Client:
    transport: Transport
    channel_client: channel.Client

    def __init__(self, transport: Transport, channel_client: channel.Client):
        self.transport = transport
        self.channel_client = channel_client

    def new_writer(self, keys: list[str]) -> writer.Numpy:
        core = writer.Core(transport=self.transport.stream)
        npw = writer.Numpy(core=core, channel_client=self.channel_client)
        npw.open(keys)
        return npw

    def new_iterator(self, keys: list[str], tr: telem.TimeRange, aggregate: bool = False) -> iterator.Numpy:
        npi = iterator.Numpy(transport=self.transport.stream,
                             channel_client=self.channel_client,
                             aggregate=aggregate)
        npi.open(keys, tr)
        return npi

    def write(self, to: str, data: np.ndarray, start: telem.UnparsedTimeStamp):
        _writer = self.new_writer([to])
        try:
            _writer.write(to, data, start)
        finally:
            _writer.close()

    def read(self, from_: str, tr: telem.TimeRange) -> np.ndarray:
        seg = self.read_seg(from_, tr)
        return seg.data

    def read_seg(self, from_: str, tr: telem.TimeRange) -> NumpySegment:
        _iterator = self.new_iterator([from_], tr, aggregate=True)
        seg = None
        try:
            _iterator.first()
            while _iterator.next():
                pass
            seg = _iterator.value[from_]
        finally:
            _iterator.close()
        return seg



from arya.transport import Transport
from . import segment
from . import channel
from freighter.endpoint import Endpoint


class Client:
    transport: Transport
    data: segment.Client
    channel: channel.Client

    def __init__(
            self,
            host: str,
            port: int,
    ):
        self.transport = Transport(Endpoint(host, port))
        self.channel = channel.Client(self.transport)
        self.data = segment.Client(self.transport, self.channel)



from typing import Protocol

import arya.errors
from arya import telem
from arya.segment import NumpySegment
from arya.telem.numpy import to_numpy_type


class Validator(Protocol):
    def validate(self, segment: NumpySegment) -> None:
        ...


class ScalarType:
    def validate(self, seg: NumpySegment) -> None:
        npt = to_numpy_type(seg.channel.data_type)
        if npt is None:
            raise arya.errors.ValidationError(
                f"Channel data type {seg.channel.data_type} is not supported"
            )

        if seg.data.dtype != npt:
            raise arya.errors.ValidationError(
                f"Expected data type {npt}, got {seg.data.dtype}"
            )

        if seg.data.ndim != 1:
            raise arya.errors.ValidationError(f"Expected 1D array, got {seg.data.ndim}")


class Contiguity:
    high_water_marks: dict[str, telem.TimeStamp]
    allow_no_high_water_mark: bool = False
    allow_overlap: bool = False
    allow_gap: bool = False

    def __init__(
        self,
        high_water_marks: dict[str, telem.TimeStamp],
        allow_no_high_water_mark: bool = False,
        allow_overlap: bool = False,
        allow_gap: bool = False,
    ) -> None:
        self.allow_no_high_water_mark = allow_no_high_water_mark
        self.allow_overlap = allow_overlap
        self.allow_gap = allow_gap
        self.high_water_marks = high_water_marks

    def validate(self, seg: NumpySegment) -> None:
        hwm = self._get_high_water_mark(seg.channel.key)
        if hwm is not None:
            self._enforce_no_overlap(hwm, seg)
            self._enforce_no_gap(hwm, seg)
        self._update_high_water_mark(seg)

    def _enforce_no_overlap(self, hwm: telem.TimeStamp, seg: NumpySegment) -> None:
        if self.allow_overlap:
            return
        if seg.start.before(hwm):
            raise arya.errors.ContiguityError(
                f"Next segment start ({seg.start}) is before the previous segments end ({hwm})"
            )

    def _enforce_no_gap(self, hwm: telem.TimeStamp, seg: NumpySegment) -> None:
        if self.allow_gap:
            return
        if seg.start != hwm:
            raise arya.errors.ContiguityError(
                f"Next segment start ({seg.start}) is not equal to the previous segments end ({hwm})"
            )

    def _update_high_water_mark(self, seg: NumpySegment) -> None:
        self.high_water_marks[seg.channel.key] = seg.end

    def _get_high_water_mark(self, channel_key: str) -> telem.TimeStamp:
        hwm = self.high_water_marks.get(channel_key, None)
        if hwm is None and not self.allow_no_high_water_mark:
            raise arya.errors.UnexpectedError(
                f"No high water mark for channel: {channel_key}"
            )
        return hwm



from arya.channel import Channel
from arya.segment import NumpySegment, BinarySegment
import numpy as np


class NumpyEncoderDecoder:
    @staticmethod
    def encode(segment: NumpySegment) -> BinarySegment:
        return BinarySegment(
            channel_key=segment.channel.key,
            start=segment.start,
            data=segment.data.tobytes(),
        )

    @staticmethod
    def decode(channel: Channel, segment: BinarySegment) -> NumpySegment:
        return NumpySegment(
            channel=channel,
            start=segment.start,
            data=np.frombuffer(segment.data, dtype=channel.numpy_type),
        )



from __future__ import annotations

from dataclasses import dataclass

from arya import telem
from arya.errors import UnexpectedError
from arya.telem.numpy import to_numpy_type

import numpy as np


@dataclass
class Channel:
    key: str
    name: str
    node_id: int
    rate: telem.Rate
    data_type: telem.DataType
    density: telem.Density

    def __init__(
            self,
            name: str = "",
            node_id: int = 0,
            rate: telem.UnparsedRate = telem.Rate(0),
            data_type: telem.UnparsedDataType = telem.DATA_TYPE_UNKNOWN,
            density: telem.UnparsedDensity = telem.Density(0),
            key: str = "",
    ):
        self.name = name
        self.node_id = node_id
        self.rate = telem.Rate(rate)
        self.data_type = telem.DataType(data_type)
        self.density = telem.Density(density)
        self.key = key

    @property
    def numpy_type(self) -> np.ScalarType:
        npt = to_numpy_type(self.data_type)
        if npt is None:
            raise UnexpectedError(
                f"Cannot convert data type {self.data_type} to numpy type"
            )
        return npt



import asyncio

from dataclasses import dataclass

import numpy as np

import arya.errors
from . import BinarySegment, NumpySegment
import freighter

from . import validate
from .. import telem
from .. import channel
from .splitter import Splitter
from .encoder import NumpyEncoderDecoder

_ENDPOINT = "/segment/write"


@dataclass
class Request:
    open_keys: list[str]
    segments: list[BinarySegment]


@dataclass
class Response:
    ack: bool
    error: freighter.ErrorPayload


def _response_factory() -> Response:
    return Response(False, freighter.ErrorPayload(None, None))


class BaseCore:
    keys: list[str]

    def _ack_open(self, res: Response | None, exc: Exception | None):
        if exc is not None:
            raise exc
        assert res is not None
        if not res.ack:
            raise arya.errors.UnexpectedError(
                "Writer failed to positively acknowledge open request. This is a bug"
                + "please report it."
            )


class AsyncCore(BaseCore):
    transport: freighter.AsyncStreamClient
    stream: freighter.AsyncStream[Request, Response]
    responses: asyncio.Task[Exception | None]

    def __init__(self, transport: freighter.AsyncStreamClient) -> None:
        self.transport = transport

    async def open(self, keys: list[str]):
        self.stream = await self.transport.stream(_ENDPOINT, Request, _response_factory)
        await self.stream.send(Request(keys, []))
        res, err = await self.stream.receive()
        self._ack_open(res, err)
        self.responses = asyncio.create_task(self.receive_errors())

    async def write(self, segments: list[BinarySegment]) -> bool:
        if self.responses.done():
            return False
        err = await self.stream.send(Request([], segments))
        if err is not None:
            raise err
        return True

    async def close(self):
        await self.stream.close_send()
        err = await self.responses
        assert err is not None
        if not isinstance(err, freighter.EOF):
            raise err

    async def receive_errors(self) -> Exception | None:
        res, exc = await self.stream.receive()
        if exc is None:
            assert res is not None
            exc = freighter.errors.decode(res.error)
        return exc


class Core(BaseCore):
    transport: freighter.StreamClient
    stream: freighter.Stream[Request, Response]

    def __init__(self, transport: freighter.StreamClient) -> None:
        self.transport = transport

    def open(self, keys: list[str]):
        self.stream = self.transport.stream(_ENDPOINT, Request, _response_factory)
        self.stream.send(Request(keys, []))
        res, err = self.stream.receive()
        self._ack_open(res, err)

    def write(self, segments: list[BinarySegment]) -> bool:
        if self.stream.received():
            return False
        err = self.stream.send(Request([], segments))
        if err is not None:
            raise err
        return True

    def close(self):
        self.stream.close_send()
        res, err = self.stream.receive()
        if err is None:
            err = freighter.errors.decode(res.error)
        if not isinstance(err, freighter.EOF):
            raise err


class Numpy:
    core: Core
    channel_client: channel.Client
    validators: list[validate.Validator]
    encoder: NumpyEncoderDecoder
    channels: channel.Registry
    splitter: Splitter

    def __init__(
        self,
        core: Core,
        channel_client: channel.Client,
    ) -> None:
        self.channel_client = channel_client
        self.core = core
        self.validators = [
            validate.ScalarType(),
            validate.Contiguity(dict(), allow_no_high_water_mark=True),
        ]
        self.encoder = NumpyEncoderDecoder()
        self.splitter = Splitter(threshold=telem.Size(4e6))

    def open(self, keys: list[str]):
        self.channels = channel.Registry(self.channel_client.retrieve(keys))
        self.core.open(keys)

    def write(self, to: str, data: np.ndarray, start: telem.UnparsedTimeStamp) -> bool:
        ch = self.channels.get(to)
        if ch is None:
            raise arya.errors.QueryError(f"channel with key {to} not found")
        seg = NumpySegment(ch, telem.TimeStamp(start), data)
        for val in self.validators:
            val.validate(seg)
        encoded = self.encoder.encode(seg).sugar(ch)
        split = self.splitter.split(encoded)
        return self.core.write([seg.desugar() for seg in split])

    def close(self):
        self.core.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['synnax', 'synnax.channel', 'synnax.segment', 'synnax.telem']

package_data = \
{'': ['*']}

install_requires = \
['mypy>=0.971,<0.972',
 'pandas>=1.4.3,<2.0.0',
 'synnax-freighter>=0.1.0,<0.2.0',
 'websockets>=10.3,<11.0']

setup_kwargs = {
    'name': 'synnax',
    'version': '0.1.0',
    'description': 'Synnax Client Library',
    'long_description': None,
    'author': 'emiliano bonilla',
    'author_email': 'emilbon99@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://synnaxlabs.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
